/***************************************************************************
 *   Copyright (C) 2005 by Peter Müller   *
 *   pmueller@cs.tu-berlin.de   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef VEC_H
#define VEC_H

#include <cmath>
using namespace std;

#include <math.h>

class Vec2 {
public:
	union {
		float data[2];
		struct {
			float x,y;
		};
	};
	
	Vec2() : x(0.0f), y(0.0f) { };
	Vec2(float _x, float _y) : x(_x), y(_y) {}
	Vec2(const Vec2 & b) : x(b.x), y(b.y) {}
	virtual ~Vec2() {}
	
	bool operator==(const Vec2 & b) const;
	bool operator!=(const Vec2 & b) const;
	
	void operator+=(const Vec2 & b);
	void operator-=(const Vec2 & b);
	void operator*=(float b);
	void operator/=(float b);
			
	Vec2 operator+(const Vec2 & b) const;
	Vec2 operator-(const Vec2 & b) const;
	Vec2 operator-() const;
	Vec2 operator*(float b) const;
	Vec2 operator/(float b) const;
	
	float dot(const Vec2 &b) const;
	float getLength() const;
	float getSquareLength() const;
	
	void normalize();
	
	static const Vec2 ZERO;
	static const Vec2 X;
	static const Vec2 Y;
};

class Vec3 {
public:
	union {
		float data[3];
		struct {
			float x,y,z;
		};
	};
	
	Vec3() : x(0.0f), y(0.0f), z(0.0f) { };
	Vec3(float _x, float _y, float _z) : x(_x), y(_y), z(_z) {}
	Vec3(const Vec3 & b) : x(b.x), y(b.y), z(b.z) {}
	~Vec3() {}
	
	bool operator==(const Vec3 & b) const;
	bool operator!=(const Vec3 & b) const;
	
	void operator+=(const Vec3 & b);
	void operator-=(const Vec3 & b);
	void operator*=(float b);
	void operator/=(float b);
		
	Vec3 operator+(const Vec3 & b) const;
	Vec3 operator-(const Vec3 & b) const;
	Vec3 operator-() const;
	Vec3 operator*(float b) const;
	Vec3 operator/(float b) const;
	
	float dot(const Vec3 &b) const;
	Vec3  cross(const Vec3 & b) const;
	float getLength() const;
	float getSquareLength() const;
	
	void normalize();
	
	// random unit length vector
	static Vec3 randomUnit();
	// normal must be normalized
	static Vec3 randomUnitOnPlane(const Vec3& normal);
	// reflect vector on plane with normal
	// normal must be normalized
	void reflect(const Vec3 & normal);
	
	static const Vec3 ZERO;
	static const Vec3 X;
	static const Vec3 Y;
	static const Vec3 Z;
};

class Vec4 {
public:
	union {
		float data[4];
		struct {
			float x,y,z,w;
		};
	};
	
	Vec4() : x(0.0f), y(0.0f), z(0.0f), w(0.0f) { };
	Vec4(float _x, float _y, float _z, float _w) : x(_x), y(_y), z(_z), w(_w) {}
	Vec4(const Vec4 & b) : x(b.x), y(b.y), z(b.z), w(b.w) {}
	~Vec4() {}
	
	bool operator==(const Vec4 & b) const;
	bool operator!=(const Vec4 & b) const;
	
	void operator+=(const Vec4 & b);
	void operator-=(const Vec4 & b);
	void operator*=(float b);
	void operator/=(float b);
		
	Vec4 operator+(const Vec4 & b) const;
	Vec4 operator-(const Vec4 & b) const;
	Vec4 operator-() const;
	Vec4 operator*(float b) const;
	Vec4 operator/(float b) const;
	
	float dot(const Vec4 &b) const;
	float getLength() const;
	float getSquareLength() const;
	
	void normalize();
	
	static const Vec4 ZERO;
	static const Vec4 X;
	static const Vec4 Y;
	static const Vec4 Z;
	static const Vec4 W;
};


// ============================== inlines ==================================

// ===============================  Vec2  ==================================

bool Vec2::operator==(const Vec2 & b) const {
	return (x == b.x) && (y == b.y);
}

bool Vec2::operator!=(const Vec2 & b) const {
	return (x != b.x) || (y != b.y);
}


void Vec2::operator+=(const Vec2 & b) {
	x += b.x;
	y += b.y;
}

void Vec2::operator-=(const Vec2 & b) {
	x -= b.x;
	y -= b.y;	
}

void Vec2::operator*=(float b) {
	x *= b;
	y *= b;
}

void Vec2::operator/=(float b) {
	x /= b;
	y /= b;
}

Vec2 Vec2::operator+(const Vec2 & b) const {
	return Vec2(x + b.x, y + b.y);
}

Vec2 Vec2::operator-(const Vec2 & b) const {
	return Vec2(x - b.x, y - b.y);
}

Vec2 Vec2::operator-() const {
	return Vec2(-x, -y);
}

Vec2 Vec2::operator*(float b) const {
	return Vec2(x*b, y*b);
}
Vec2 Vec2::operator/(float b) const {
	return Vec2(x/b, y/b);
}

float Vec2::dot(const Vec2 &b) const {
	return x*b.x + y*b.y;
}

float Vec2::getLength() const {
	return sqrtf(x*x + y*y);
}

float Vec2::getSquareLength() const {
	return x*x + y*y;
}

void Vec2::normalize() {
	float l = getLength();
	x /= l;
	y /= l;
}

// ===============================  Vec3  ==================================

bool Vec3::operator==(const Vec3 & b) const {
	return (x == b.x) && (y == b.y) && (z == b.z);
}

bool Vec3::operator!=(const Vec3 & b) const {
	return (x != b.x) || (y != b.y) || (z != b.z);
}

void Vec3::operator+=(const Vec3 & b) {
	x += b.x;
	y += b.y;
	z += b.z;
}

void Vec3::operator-=(const Vec3 & b) {
	x -= b.x;
	y -= b.y;
	z -= b.z;
}

void Vec3::operator*=(float b) {
	x *= b;
	y *= b;
	z *= b;
}

void Vec3::operator/=(float b) {
	x /= b;
	y /= b;
	z /= b;
}

Vec3 Vec3::operator+(const Vec3 & b) const {
	return Vec3(x + b.x, y + b.y, z + b.z);
}

Vec3 Vec3::operator-(const Vec3 & b) const {
	return Vec3(x - b.x, y - b.y, z - b.z);
}

Vec3 Vec3::operator-() const {
	return Vec3(-x, -y, -z);
}

Vec3 Vec3::operator*(float b) const {
	return Vec3(x*b, y*b, z*b);
}

Vec3 Vec3::operator/(float b) const {
	return Vec3(x/b, y/b, z/b);
}

float Vec3::dot(const Vec3 &b) const {
	return x*b.x + y*b.y + z*b.z;
}

Vec3 Vec3::cross(const Vec3 &b) const {
	return Vec3(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);
}

float Vec3::getLength() const {
	return sqrtf(x*x + y*y + z*z);
}

float Vec3::getSquareLength() const {
	return x*x + y*y + z*z;
}

void Vec3::normalize() {
	float a = 1 / getLength();
	x *= a;
	y *= a;
	z *= a;
}

// ===============================  Vec4  ==================================

bool Vec4::operator==(const Vec4 & b) const {
	return (x == b.x) && (y == b.y) && (z == b.z) && (w == b.w);
}

bool Vec4::operator!=(const Vec4 & b) const {
	return (x != b.x) || (y != b.y) || (z != b.z) || (w != b.w);
}

void Vec4::operator+=(const Vec4 & b) {
	x += b.x;
	y += b.y;
	z += b.z;
	w += b.w;
}

void Vec4::operator-=(const Vec4 & b) {
	x -= b.x;
	y -= b.y;
	z -= b.z;
	w -= b.w;
}

void Vec4::operator*=(float b) {
	x *= b;
	y *= b;
	z *= b;
	w *= b;
}

void Vec4::operator/=(float b) {
	x /= b;
	y /= b;
	z /= b;
	w /= b;
}

Vec4 Vec4::operator+(const Vec4 & b) const {
	return Vec4(x + b.x, y + b.y, z + b.z, w + b.w);
}

Vec4 Vec4::operator-(const Vec4 & b) const {
	return Vec4(x - b.x, y - b.y, z - b.z, w - b.w);
}

Vec4 Vec4::operator-() const {
	return Vec4(-x, -y, -z, -w);
}

Vec4 Vec4::operator*(float b) const {
	return Vec4(x*b, y*b, z*b, w*b);
}

Vec4 Vec4::operator/(float b) const {
	return Vec4(x/b, y/b, z/b, w/b);
}

float Vec4::dot(const Vec4 &b) const {
	return x*b.x + y*b.y + z*b.z + w*b.w;
}

float Vec4::getLength() const {
	return sqrtf(x*x + y*y + z*z + w*w);
}

float Vec4::getSquareLength() const {
	return x*x + y*y + z*z + w*w;
}

void Vec4::normalize() {
	float l = getLength();
	x /= l;
	y /= l;
	z /= l;
	w /= l;
}

#endif
